
# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from shutil import copy

setup(
    name='vaitrace_py',
    version='1.3',
    description="",
    py_modules=["__init__", "vaitrace_py", "tracepoint"],
    python_requires='>=3.5',
    zip_safe=False,
)
