/*
 * Copyright 2019 Xilinx Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "vart/assistant/tensor_buffer_allocator.hpp"

// defined in tensor_buffer_allocator_imp.cpp
extern "C" vart::assistant::TensorBufferAllocator*
create_tensor_buffer_allocator(const xir::Attrs* attrs);

namespace vart {
namespace assistant {
TensorBufferAllocator::TensorBufferAllocator() {}

std::unique_ptr<TensorBufferAllocator> TensorBufferAllocator::create(
    const xir::Attrs* attrs) {
  return std::unique_ptr<TensorBufferAllocator>(
      create_tensor_buffer_allocator(attrs));
}
}  // namespace assistant
}  // namespace vart
